@@ -68,7 +68,7 @@
  *		Note. under this mechanism, any data item that undergoes
  *		relocation and is then further modified during the execution of
  *		the image before dynodump(3x) is called will lose the
- *		modification that occured during the applications execution.
+ *		modification that occurred during the applications execution.
  *
  * N.B. The above commentary is not quite correct in the flags have been hardwired
  *      to RTLD_SAVREL.
